{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Multiple viewer widget\n\nThis is an example of how to have more than one viewer in the same napari window.\nAdditional viewers state will be synchronized with the main viewer.\nSwitching to 3D display will only impact the main viewer.\n\nThis example also contains the option to enable cross that will be moved to the\ncurrent dims point (`viewer.dims.point`).\n\n.. tags:: gui\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "from copy import deepcopy\n\nimport numpy as np\nfrom qtpy.QtCore import Qt\nfrom qtpy.QtWidgets import (\n    QCheckBox,\n    QDoubleSpinBox,\n    QPushButton,\n    QSplitter,\n    QTabWidget,\n    QVBoxLayout,\n    QWidget,\n)\nfrom superqt.utils import qthrottled\n\nimport napari\nfrom napari.components.viewer_model import ViewerModel\nfrom napari.layers import Labels, Layer, Vectors\nfrom napari.qt import QtViewer\nfrom napari.utils.action_manager import action_manager\nfrom napari.utils.events.event import WarningEmitter\nfrom napari.utils.notifications import show_info\n\n\ndef copy_layer(layer: Layer, name: str = ''):\n    res_layer = Layer.create(*layer.as_layer_data_tuple())\n    res_layer.metadata['viewer_name'] = name\n    return res_layer\n\n\ndef get_property_names(layer: Layer):\n    klass = layer.__class__\n    res = []\n    for event_name, event_emitter in layer.events.emitters.items():\n        if isinstance(event_emitter, WarningEmitter):\n            continue\n        if event_name in ('thumbnail', 'name'):\n            continue\n        if (\n            isinstance(getattr(klass, event_name, None), property)\n            and getattr(klass, event_name).fset is not None\n        ):\n            res.append(event_name)\n    return res\n\n\ndef center_cross_on_mouse(\n    viewer_model: napari.components.viewer_model.ViewerModel,\n):\n    \"\"\"move the cross to the mouse position\"\"\"\n\n    if not getattr(viewer_model, 'mouse_over_canvas', True):\n        # There is no way for napari 0.4.15 to check if mouse is over sending canvas.\n        show_info(\n            'Mouse is not over the canvas. You may need to click on the canvas.'\n        )\n        return\n\n    viewer_model.dims.current_step = tuple(\n        np.round(\n            [\n                max(min_, min(p, max_)) / step\n                for p, (min_, max_, step) in zip(\n                    viewer_model.cursor.position, viewer_model.dims.range, strict=False\n                )\n            ]\n        ).astype(int)\n    )\n\n\naction_manager.register_action(\n    name='napari:move_point',\n    command=center_cross_on_mouse,\n    description='Move dims point to mouse position',\n    keymapprovider=ViewerModel,\n)\n\naction_manager.bind_shortcut('napari:move_point', 'C')\n\n\nclass own_partial:\n    \"\"\"\n    Workaround for deepcopy not copying partial functions\n    (Qt widgets are not serializable)\n    \"\"\"\n\n    def __init__(self, func, *args, **kwargs) -> None:\n        self.func = func\n        self.args = args\n        self.kwargs = kwargs\n\n    def __call__(self, *args, **kwargs):\n        return self.func(*(self.args + args), **{**self.kwargs, **kwargs})\n\n    def __deepcopy__(self, memodict=None):\n        if memodict is None:\n            memodict = {}\n        return own_partial(\n            self.func,\n            *deepcopy(self.args, memodict),\n            **deepcopy(self.kwargs, memodict),\n        )\n\n\nclass QtViewerWrap(QtViewer):\n    def __init__(self, main_viewer, *args, **kwargs) -> None:\n        super().__init__(*args, **kwargs)\n        self.main_viewer = main_viewer\n\n    def _qt_open(\n        self,\n        filenames: list,\n        stack: bool,\n        plugin: str | None = None,\n        layer_type: str | None = None,\n        **kwargs,\n    ):\n        \"\"\"for drag and drop open files\"\"\"\n        self.main_viewer.window._qt_viewer._qt_open(\n            filenames, stack, plugin, layer_type, **kwargs\n        )\n\n\nclass CrossWidget(QCheckBox):\n    \"\"\"\n    Widget to control the layer representing cross.\n    Because of the performance reasons,\n    the update of cross is throttled\n    \"\"\"\n\n    def __init__(self, viewer: napari.Viewer) -> None:\n        super().__init__('Add cross layer')\n        self.viewer = viewer\n        self.setChecked(False)\n        self.stateChanged.connect(self._update_cross_visibility)\n        self.layer = None\n        self.viewer.dims.events.order.connect(self.update_cross)\n        self.viewer.dims.events.ndim.connect(self._update_ndim)\n        self.viewer.dims.events.current_step.connect(self.update_cross)\n        self._extent = None\n\n        self._update_extent()\n        self.viewer.dims.events.connect(self._update_extent)\n\n    @qthrottled(leading=False)\n    def _update_extent(self):\n        \"\"\"\n        Calculate the extent of the data.\n\n        Ignores the layer with cross itself in calculating the extent.\n        \"\"\"\n        layers = [\n            layer\n            for layer in self.viewer.layers\n            if layer is not self.layer\n        ]\n        self._extent = self.viewer.layers.get_extent(layers)\n        self.update_cross()\n\n    def _update_ndim(self, event):\n        if self.layer in self.viewer.layers:\n            self.viewer.layers.remove(self.layer)\n        self.layer = Vectors(name='.cross', ndim=event.value)\n        self.layer.edge_width = 1.5\n        self.update_cross()\n\n    def _update_cross_visibility(self, state):\n        if state:\n            self.viewer.layers.append(self.layer)\n        else:\n            self.viewer.layers.remove(self.layer)\n        self.update_cross()\n\n    def update_cross(self):\n        if self.layer not in self.viewer.layers:\n            return\n\n        point = self.viewer.dims.current_step\n        vec = []\n        for i, (lower, upper) in enumerate(self._extent.world.T):\n            if (upper - lower) / self._extent.step[i] == 1:\n                continue\n            point1 = list(point)\n            point1[i] = (lower + self._extent.step[i] / 2) / self._extent.step[\n                i\n            ]\n            point2 = [0 for _ in point]\n            point2[i] = (upper - lower) / self._extent.step[i]\n            vec.append((point1, point2))\n        if np.any(self.layer.scale != self._extent.step):\n            self.layer.scale = self._extent.step\n        self.layer.data = vec\n\n\nclass ExampleWidget(QWidget):\n    \"\"\"\n    Example widget showcasing how to place additional widgets to the right\n    of the additional viewers.\n    \"\"\"\n\n    def __init__(self) -> None:\n        super().__init__()\n        self.btn = QPushButton('Perform action')\n        self.spin = QDoubleSpinBox()\n        layout = QVBoxLayout()\n        layout.addWidget(self.spin)\n        layout.addWidget(self.btn)\n        layout.addStretch(1)\n        self.setLayout(layout)\n\n\nclass MultipleViewerWidget(QSplitter):\n    \"\"\"The main widget of the example.\"\"\"\n\n    def __init__(self, viewer: napari.Viewer) -> None:\n        super().__init__()\n        self.viewer = viewer\n        self.viewer_model1 = ViewerModel(title='model1')\n        self.viewer_model2 = ViewerModel(title='model2')\n        self._block = False\n        self.qt_viewer1 = QtViewerWrap(viewer, self.viewer_model1)\n        self.qt_viewer2 = QtViewerWrap(viewer, self.viewer_model2)\n        self.tab_widget = QTabWidget()\n        w1 = ExampleWidget()\n        w2 = ExampleWidget()\n        self.tab_widget.addTab(w1, 'Sample 1')\n        self.tab_widget.addTab(w2, 'Sample 2')\n        viewer_splitter = QSplitter()\n        viewer_splitter.setOrientation(Qt.Orientation.Vertical)\n        viewer_splitter.addWidget(self.qt_viewer1)\n        viewer_splitter.addWidget(self.qt_viewer2)\n        viewer_splitter.setContentsMargins(0, 0, 0, 0)\n\n        self.addWidget(viewer_splitter)\n        self.addWidget(self.tab_widget)\n\n        self.viewer.layers.events.inserted.connect(self._layer_added)\n        self.viewer.layers.events.removed.connect(self._layer_removed)\n        self.viewer.layers.events.moved.connect(self._layer_moved)\n        self.viewer.layers.selection.events.active.connect(\n            self._layer_selection_changed\n        )\n        self.viewer.dims.events.current_step.connect(self._point_update)\n        self.viewer_model1.dims.events.current_step.connect(self._point_update)\n        self.viewer_model2.dims.events.current_step.connect(self._point_update)\n        self.viewer.dims.events.order.connect(self._order_update)\n        self.viewer.events.reset_view.connect(self._reset_view)\n        self.viewer_model1.events.status.connect(self._status_update)\n        self.viewer_model2.events.status.connect(self._status_update)\n\n    def _status_update(self, event):\n        self.viewer.status = event.value\n\n    def _reset_view(self):\n        self.viewer_model1.reset_view()\n        self.viewer_model2.reset_view()\n\n    def _layer_selection_changed(self, event):\n        \"\"\"\n        update of current active layer\n        \"\"\"\n        if self._block:\n            return\n\n        if event.value is None:\n            self.viewer_model1.layers.selection.active = None\n            self.viewer_model2.layers.selection.active = None\n            return\n\n        self.viewer_model1.layers.selection.active = self.viewer_model1.layers[\n            event.value.name\n        ]\n        self.viewer_model2.layers.selection.active = self.viewer_model2.layers[\n            event.value.name\n        ]\n\n    def _point_update(self, event):\n        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:\n            if model.dims is event.source:\n                continue\n            if len(self.viewer.layers) != len(model.layers):\n                continue\n            model.dims.current_step = event.value\n\n    def _order_update(self):\n        order = list(self.viewer.dims.order)\n        if len(order) <= 2:\n            self.viewer_model1.dims.order = order\n            self.viewer_model2.dims.order = order\n            return\n\n        order[-3:] = order[-2], order[-3], order[-1]\n        self.viewer_model1.dims.order = tuple(order)\n        order = list(self.viewer.dims.order)\n        order[-3:] = order[-1], order[-2], order[-3]\n        self.viewer_model2.dims.order = tuple(order)\n\n    def _layer_added(self, event):\n        \"\"\"add layer to additional viewers and connect all required events\"\"\"\n        self.viewer_model1.layers.insert(\n            event.index, copy_layer(event.value, 'model1')\n        )\n        self.viewer_model2.layers.insert(\n            event.index, copy_layer(event.value, 'model2')\n        )\n        for name in get_property_names(event.value):\n            getattr(event.value.events, name).connect(\n                own_partial(self._property_sync, name)\n            )\n\n        if isinstance(event.value, Labels):\n            event.value.events.set_data.connect(self._set_data_refresh)\n            event.value.events.labels_update.connect(self._set_data_refresh)\n            self.viewer_model1.layers[\n                event.value.name\n            ].events.set_data.connect(self._set_data_refresh)\n            self.viewer_model2.layers[\n                event.value.name\n            ].events.set_data.connect(self._set_data_refresh)\n            event.value.events.labels_update.connect(self._set_data_refresh)\n            self.viewer_model1.layers[\n                event.value.name\n            ].events.labels_update.connect(self._set_data_refresh)\n            self.viewer_model2.layers[\n                event.value.name\n            ].events.labels_update.connect(self._set_data_refresh)\n        if event.value.name != '.cross':\n            self.viewer_model1.layers[event.value.name].events.data.connect(\n                self._sync_data\n            )\n            self.viewer_model2.layers[event.value.name].events.data.connect(\n                self._sync_data\n            )\n\n        event.value.events.name.connect(self._sync_name)\n\n        self._order_update()\n\n    def _sync_name(self, event):\n        \"\"\"sync name of layers\"\"\"\n        index = self.viewer.layers.index(event.source)\n        self.viewer_model1.layers[index].name = event.source.name\n        self.viewer_model2.layers[index].name = event.source.name\n\n    def _sync_data(self, event):\n        \"\"\"sync data modification from additional viewers\"\"\"\n        if self._block:\n            return\n        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:\n            layer = model.layers[event.source.name]\n            if layer is event.source:\n                continue\n            try:\n                self._block = True\n                layer.data = event.source.data\n            finally:\n                self._block = False\n\n    def _set_data_refresh(self, event):\n        \"\"\"\n        synchronize data refresh between layers\n        \"\"\"\n        if self._block:\n            return\n        for model in [self.viewer, self.viewer_model1, self.viewer_model2]:\n            layer = model.layers[event.source.name]\n            if layer is event.source:\n                continue\n            try:\n                self._block = True\n                layer.refresh()\n            finally:\n                self._block = False\n\n    def _layer_removed(self, event):\n        \"\"\"remove layer in all viewers\"\"\"\n        self.viewer_model1.layers.pop(event.index)\n        self.viewer_model2.layers.pop(event.index)\n\n    def _layer_moved(self, event):\n        \"\"\"update order of layers\"\"\"\n        dest_index = (\n            event.new_index\n            if event.new_index < event.index\n            else event.new_index + 1\n        )\n        self.viewer_model1.layers.move(event.index, dest_index)\n        self.viewer_model2.layers.move(event.index, dest_index)\n\n    def _property_sync(self, name, event):\n        \"\"\"Sync layers properties (except the name)\"\"\"\n        if event.source not in self.viewer.layers:\n            return\n        try:\n            self._block = True\n            setattr(\n                self.viewer_model1.layers[event.source.name],\n                name,\n                getattr(event.source, name),\n            )\n            setattr(\n                self.viewer_model2.layers[event.source.name],\n                name,\n                getattr(event.source, name),\n            )\n        finally:\n            self._block = False\n\n\nif __name__ == '__main__':\n    from qtpy import QtWidgets\n    QtWidgets.QApplication.setAttribute(Qt.ApplicationAttribute.AA_ShareOpenGLContexts)\n    # above two lines are needed to allow undocking the widget with\n    # additional viewers\n    view = napari.Viewer()\n    dock_widget = MultipleViewerWidget(view)\n    cross = CrossWidget(view)\n\n    view.window.add_dock_widget(dock_widget, name='Sample')\n    view.window.add_dock_widget(cross, name='Cross', area='left')\n\n    view.open_sample('napari', 'cells3d')\n\n    napari.run()"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.10.16"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}